import os
from bears.codeclone_detection.CountVector import CountVector
from bears.codeclone_detection.ClangCountingConditions import (
    is_reference,
    get_identifier_name,
    is_literal,
    is_function_declaration)
from coalib.bearlib.parsing.clang.cindex import Cursor, Index


def get_include_paths(file_path, setting_path):
    """
    Creates a list of include paths that likely resolve all includes.

    :param file_path:    The path to the file to analyze.
    :param setting_path: The path of the coafile.
    :return:             All directories that lie in between the common subpath
                         of those two and the file_path.
    """
    path = os.path.dirname(os.path.commonprefix(
        [os.path.abspath(file_path), os.path.abspath(setting_path)]))
    result = [path]
    for directory in file_path[len(path)+1:].split(os.path.sep)[:-1]:
        path = os.path.join(path, directory)
        result.append(path)

    return result


class ClangCountVectorCreator:
    """
    This object uses clang to create a count vector for each function for given
    counting conditions. The counting conditions are clang specific and they
    are called like this:

      condition(stack)

    While stack is a stack (i.e. list) holding a tuple holding the parent
    cursors and the child number. (E.g. if a cursor is the third child of
    its parent its child number is two, counted from zero.)

    The ClangCountVectorCreator will only count variables local to each
    function.
    """
    def __init__(self,
                 conditions=None,
                 weightings=None,
                 definition_path=None):
        """
        Creates a new ClangCountVectorCreator.

        :param conditions:      The counting conditions as list of function
                                objects, each shall return true when getting
                                data indicating that this occurrence should
                                be counted.
        :param weightings:      Optional factors to weight counting conditions.
                                Defaults to 1 for all conditions.
        :param definition_path: section["files"].origin, i.e. the path where
                                the user defined the files. Include paths will
                                be guessed with the help of this information if
                                it is available.
        """
        self.conditions = conditions
        self.weightings = weightings
        self.definition_path = definition_path
        self.count_vectors = {}
        self.stack = []

    def count_identifier(self, identifier):
        if identifier not in self.count_vectors:
            self.count_vectors[identifier] = (
                CountVector(identifier, self.conditions, self.weightings))

        self.count_vectors[identifier].count_reference(self.stack)

    def _get_vector_for_function(self, cursor, child_num=0):
        """
        Creates a CountVector object for the given cursor.

        Note: this function uses self.count_vectors for storing its results.
        This is done knowingly because passing back and forth mutable objects
        is not nice and yields in bigger complexity IMHO.

        This function creates a CountVector object for all variables found in
        self.local_vars and in the tree elements below the given one, stores it
        in self.count_vectors.

        :param cursor: Clang cursor to iterate over.
        """
        assert isinstance(cursor, Cursor)
        self.stack.append((cursor, child_num))

        if is_reference(cursor):
            self.count_identifier(get_identifier_name(cursor))
        if is_literal(cursor):
            tokens = list(cursor.get_tokens())
            if tokens:
                # Mangle constants with $ (-> no valid C identifier), first
                # token is the constant, semicolon and similar things may
                # follow, don't want them
                self.count_identifier("#" + tokens[0].spelling.decode())

        for i, child in enumerate(cursor.get_children()):
            self._get_vector_for_function(child, i)

        self.stack.pop()

    def _get_vectors_for_cursor(self, cursor, filename):
        """
        Maps all functions in/under the given cursor to their count vectors
        if they are defined in the given file.

        :param cursor:   The cursor to traverse.
        :param filename: Absolute path to the file.
        :return:         The dictionary holding CountVectors for all variables
                         in all functions.
        """
        assert isinstance(cursor, Cursor)
        file = cursor.location.file
        if file is not None:
            file = file.name.decode()

        if str(file) == str(filename) and is_function_declaration(cursor):
            self._get_vector_for_function(cursor)

            result = {(cursor.extent.start.line,
                       get_identifier_name(cursor)): self.count_vectors}
            # Reset local states
            self.count_vectors = {}
            self.stack = []
        else:
            result = {}
            for child in cursor.get_children():
                result.update(self._get_vectors_for_cursor(child, filename))

        return result

    def get_vectors_for_file(self, filename):
        """
        Creates a dictionary associating each function name within the given
        file with another dictionary associating each variable name (local to
        the function) with a CountVector object. Functions of included files
        will not be analyzed.

        :param filename: The path to the file to parse.
        :return:         The dictionary holding CountVectors for all variables
                         in all functions.
        """
        if self.definition_path is not None:
            include_paths = get_include_paths(filename, self.definition_path)
        else:
            include_paths = []

        args = ["-I"+path for path in include_paths]
        root = Index.create().parse(filename, args=args).cursor

        return self._get_vectors_for_cursor(root, filename)



from coalib.output.ClosableObject import ClosableObject
from coalib.output.printers.LogPrinter import LogPrinter
from coalib.output.printers.LOG_LEVEL import LOG_LEVEL


class FilePrinter(LogPrinter, ClosableObject):
    """
    This is a simple printer/logprinter that prints everything to a file. Note
    that everything will be appended.
    """
    def __init__(self,
                 filename,
                 log_level=LOG_LEVEL.WARNING,
                 timestamp_format="%X"):
        """
        Creates a new FilePrinter. If the directory of the given file doesn't
        exist or if there's any access problems, an exception will be thrown.

        :param filename: the name of the file to put the data into (string).
        """
        ClosableObject.__init__(self)
        self.file = None
        if not isinstance(filename, str):
            raise TypeError("filename must be a string.")

        LogPrinter.__init__(self,
                            timestamp_format=timestamp_format,
                            log_level=log_level)

        self.file = open(filename, 'a+')

    def _close(self):
        if self.file is not None:
            self.file.close()

    def _print(self, output, **kwargs):
        self.file.write(output)



from coalib.bearlib.spacing.SpacingHelper import SpacingHelper
from coalib.bears.LocalBear import LocalBear
from coalib.results.Result import Result
from coalib.misc.i18n import _


class LineLengthBear(LocalBear):
    # pylint: disable=unused-argument
    def run(self,
            filename,
            file,
            max_line_length: int,
            tab_width: int=SpacingHelper.DEFAULT_TAB_WIDTH):
        '''
        Yields results for all lines longer than the given maximum line length.

        :param max_line_length: Maximum number of characters for a line.
        :param tab_width: Number of spaces to show for one tab.
        '''
        spacing_helper = SpacingHelper.from_section(section=self.section)

        for line_number, line in enumerate(file):
            line = spacing_helper.replace_tabs_with_spaces(line)
            if len(line) > max_line_length + 1:
                yield Result(origin=self,
                             message=_("Line is longer than allowed.") +
                                     " ({actual} > {maximum})".format(
                                         actual=len(line),
                                         maximum=max_line_length),
                             file=filename,
                             line_nr=line_number + 1)



import shutil
try:
    # This import has side effects and is needed to make input() behave nicely
    import readline  # pylint: disable=unused-import
except ImportError: # pragma: no cover
    pass

from coalib.output.printers.ConsolePrinter import ConsolePrinter
from coalib.results.RESULT_SEVERITY import (
    RESULT_SEVERITY,
    RESULT_SEVERITY_COLORS)
from coalib.output.printers.LOG_LEVEL import LOG_LEVEL
from coalib.misc.i18n import _
from coalib.settings.Setting import Setting
from coalib.results.Result import Result
from coalib.misc.DictUtilities import inverse_dicts


STR_GET_VAL_FOR_SETTING = _("Please enter a value for the setting \"{}\" ({}) "
                            "needed by {}: ")
STR_LINE_DOESNT_EXIST = _("The line belonging to the following result "
                          "cannot be printed because it refers to a line "
                          "that doesn't seem to exist in the given file.")
STR_PROJECT_WIDE = _("Project wide:")
FILE_NAME_COLOR = "blue"
FILE_LINES_COLOR = "blue"


def format_line(line, real_nr="", sign="|", mod_nr="", symbol="", ):
    return "|{:>4}{}{:>4}|{:1}{}".format(real_nr,
                                         sign,
                                         mod_nr,
                                         symbol,
                                         line.rstrip("\n"))


def print_section_beginning(console_printer, section):
    """
    Will be called after initialization current_section in
    begin_section()

    :param console_printer: Object to print messages on the console.
    :param section:         The section that will get executed now.
    """
    console_printer.print(_("Executing section {name}...").format(
        name=section.name))


def nothing_done(console_printer):
    """
    Will be called after processing a coafile when nothing had to be done,
    i.e. no section was enabled/targeted.
    """
    console_printer.print(_("No existent section was targeted or enabled. "
                            "Nothing to do."))


def finalize(file_diff_dict, file_dict):
    """
    To be called after all results are printed on the console.

    :param file_diff_dict: A dictionary containing filenames as keys and diff
                           objects as values.
    :param file_dict:      A dictionary containing all files with filename as
                           key.
    """
    for filename in file_diff_dict:
        diff = file_diff_dict[filename]
        file_dict[filename] = diff.apply(file_dict[filename])

        # Backup original file, override old backup if needed
        shutil.copy2(filename, filename + ".orig")

        # Write new contents
        with open(filename, mode='w') as file:
            file.writelines(file_dict[filename])
    file_diff_dict.clear()


def print_lines(console_printer,
                pre_padding,
                file_dict,
                current_line,
                result_line,
                result_file):
    """
    Prints the lines between the current and the result line. If needed
    they will be shortened.

    :param console_printer: Object to print messages on the console.
    :param file_dict      : A dictionary containing all files with filename as
                            key.
    :param current_line   : The current line printed by the console_printer.
    :param result_line    : The line to which the result belongs to.
    :param result_file    : The file to which the result belongs to.
    """
    line_delta = result_line - current_line

    if line_delta > pre_padding:
        print_segregation(console_printer)

        for i in range(max(result_line - pre_padding, 1),
                       result_line + 1):
            console_printer.print(
                format_line(line=file_dict[result_file][i - 1],
                            real_nr=i,
                            mod_nr=i),
                color=FILE_LINES_COLOR)
    else:
        for i in range(1, line_delta + 1):
            console_printer.print(
                format_line(
                    line=file_dict[result_file][current_line + i - 1],
                    real_nr=current_line + i,
                    mod_nr=current_line + i),
                color=FILE_LINES_COLOR)


def print_result(console_printer,
                 log_printer,
                 section,
                 file_diff_dict,
                 result,
                 file_dict):
    """
    Prints the result to console.

    :param console_printer: Object to print messages on the console.
    :param log_printer:     Printer responsible for logging the messages.
    :param section:         Name of section to which the result belongs.
    :param file_diff_dict:  Dictionary containing filenames as keys and Diff
                            objects as values.
    :param result:          A derivative of Result.
    :param file_dict:       A dictionary containing all files with filename as
                            key.
    """
    if not isinstance(result, Result):
        log_printer.warn(_("One of the results can not be printed since it is "
                           "not a valid derivative of the coala result "
                           "class."))
        return

    console_printer.print(format_line("[{sev}] {bear}:".format(
        sev=RESULT_SEVERITY.__str__(result.severity), bear=result.origin)),
        color=RESULT_SEVERITY_COLORS[result.severity])
    console_printer.print(
        *[format_line(line) for line in result.message.split("\n")],
        delimiter="\n")

    actions = result.get_actions()
    if actions == []:
        return

    action_dict = {}
    metadata_list = []
    for action in actions:
        metadata = action.get_metadata()
        action_dict[metadata.name] = action
        metadata_list.append(metadata)

    # User can always choose no action which is guaranteed to succeed
    while not apply_action(log_printer,
                           console_printer,
                           section,
                           metadata_list,
                           action_dict,
                           result,
                           file_diff_dict,
                           file_dict):
        pass


def print_results(log_printer,
                  section,
                  result_list,
                  file_dict,
                  file_diff_dict,
                  color=True,
                  pre_padding=3):
    """
    Print all the results in a section.

    :param log_printer:    Printer responsible for logging the messages.
    :param section:        The section to which the results belong to.
    :param result_list:    List containing the results
    :param file_dict:      A dictionary containing all files with filename as
                           key.
    :param file_diff_dict: A dictionary that contains filenames as keys and
                           diff objects as values.
    :param color:          Boolean variable to print the results in color or
                           not.
    :param pre_padding:    No of lines of file to print before the result line.
                           Default value is 3.
    """
    if not isinstance(result_list, list):
        raise TypeError("result_list should be of type list")
    if not isinstance(file_dict, dict):
        raise TypeError("file_dict should be of type dict")
    if not isinstance(file_diff_dict, dict):
        raise TypeError("file_diff_dict should be of type dict")

    # We can't use None since we need line 109 be executed if file of first
    # result is None
    console_printer = ConsolePrinter(print_colored=color)
    current_file = False
    current_line = 0

    for result in sorted(result_list):
        if result.file != current_file:
            if result.file in file_dict or result.file is None:
                current_file = result.file
                current_line = 0
                console_printer.print("\n\n{}".format(current_file
                                                      if current_file is not
                                                      None
                                                      else STR_PROJECT_WIDE),
                                      color=FILE_NAME_COLOR)
            else:
                log_printer.warn(_("A result ({}) cannot be printed "
                                   "because it refers to a file that "
                                   "doesn't seem to "
                                   "exist.").format(str(result)))
                continue

        if result.line_nr is not None:
            if current_file is None:
                raise AssertionError("A result with a line_nr should also "
                                     "have a file.")
            if result.line_nr < current_line:  # pragma: no cover
                raise AssertionError("The sorting of the results doesn't "
                                     "work correctly.")
            if len(file_dict[result.file]) < result.line_nr - 1:
                console_printer.print(format_line(line=STR_LINE_DOESNT_EXIST))
            else:
                print_lines(console_printer,
                            pre_padding,
                            file_dict,
                            current_line,
                            result.line_nr,
                            result.file)
                current_line = result.line_nr

        print_result(console_printer,
                     log_printer,
                     section,
                     file_diff_dict,
                     result,
                     file_dict)


def require_setting(log_printer, setting_name, arr):
    """
    This method is responsible for prompting a user about a missing setting and
    taking its value as input from the user.

    :param log_printer:  Printer responsible for logging the messages.
    :param setting_name: Name od the setting missing
    :param arr:          a list containing a description in [0] and the name
                         of the bears who need this setting in [1] and
                         following.
    """
    if not isinstance(arr, list) or len(arr) < 2:
        log_printer.log(LOG_LEVEL.WARNING,
                        _("One of the given settings ({}) is not properly "
                          "described.").format(str(setting_name)))

        return None

    if len(arr) == 2:
        needed = arr[1]
    else:
        needed = ", ".join(arr[1:-1]) + _(" and ") + arr[-1]

    return input(STR_GET_VAL_FOR_SETTING.format(str(setting_name),
                                                str(arr[0]),
                                                needed))


def acquire_settings(log_printer, settings_names_dict):
    """
    This method prompts the user for the given settings.

    :param log_printer: Printer responsible for logging the messages.
    :param settings:    a dictionary with the settings name as key and a list
                        containing a description in [0] and the name of the
                        bears who need this setting in [1] and following.
                     Example:
    {"UseTabs": ["describes whether tabs should be used instead of spaces",
                 "SpaceConsistencyBear",
                 "SomeOtherBear"]}

    :return:            a dictionary with the settings name as key and the
                        given value as value.
    """
    if not isinstance(settings_names_dict, dict):
        raise TypeError("The settings_names_dict parameter has to be a "
                        "dictionary.")

    result = {}
    for setting_name, arr in settings_names_dict.items():
        value = require_setting(log_printer, setting_name, arr)
        if value is not None:
            result[setting_name] = value

    return result


def get_action_info(section, action):
    """
    Get all the required Settings for an action. It updates the section with
    the Settings.

    :param section: The section the action corresponds to.
    :param action:  The action to get the info for.
    :return:        Action name and the updated section.
    """
    params = action.non_optional_params

    if section is None:
        raise ValueError("section has to be intializied.")

    for param_name in params:
        if param_name not in section:
            question = format_line(
                _("Please enter a value for the parameter '{}' ({}): ")
                .format(param_name, params[param_name][0]))
            section.append(Setting(param_name, input(question)))

    return action.name, section


def choose_action(console_printer, actions):
    """
    Presents the actions available to the user and takes as input the action
    the user wants to choose.

    :param console_printer: Object to print messages on the console.
    :param actions:         Actions available to the user.
    :return:                Return choice of action of user.
    """
    console_printer.print(format_line(
        _("The following options are applicable to this result:")))

    while True:
        console_printer.print(format_line(" 0: " + _("Do nothing.")))
        for i, action in enumerate(actions):
            console_printer.print(format_line("{:>2}: {}".format(i + 1,
                                                                 action.desc)))

        try:
            line = format_line(_("Please enter the number of the action "
                                 "you want to execute. "))
            choice = int(input(line))
            if 0 <= choice <= len(actions):
                return choice
        except ValueError:
            pass

        console_printer.print(format_line(_("Please enter a valid number.")))


def print_actions(console_printer, section, actions):
    """
    Prints the given actions and lets the user choose.

    :param actions: A list of FunctionMetadata objects.
    :return:        A touple with the name member of the FunctionMetadata
                    object chosen by the user and a Section containing at
                    least all needed values for the action. If the user did
                    choose to do nothing, return (None, None).
    """
    choice = choose_action(console_printer, actions)

    if choice == 0:
        return None, None

    return get_action_info(section, actions[choice - 1])


def print_action_failed(log_printer, action_name, exception):
    """
    Prints out the information that the chosen action failed.

    :param log_printer: Printer responsible for logging the messages.
    :param action_name: The name of the action that failed.
    :param exception:   The exception with which it failed.
    """
    log_printer.log_exception("Failed to execute the action {}.".
                              format(action_name), exception)


def apply_action(log_printer,
                 console_printer,
                 section,
                 metadata_list,
                 action_dict,
                 result,
                 file_diff_dict,
                 file_dict):
    """
    Applies the action selected by the user.

    :param log_printer:     Printer responsible for logging the messages.
    :param console_printer: Object to print messages on the console.
    :param section:         Currently active section.
    :param metadata_list:   Contains metadata for all the actions.
    :param action_dict:     Contains the action names as keys and their
                            references as values.
    :param result:          Result corresponding to the actions.
    :param file_diff_dict:  If its an action which applies a patch, this
                            contains the diff of the patch to be applied to
                            the file with filename as keys.
    :param file_dict:       Dictionary with filename as keys and its contents
                            as values.
    :return:                Returns a boolean value. True if the action got
                            applied successfully, and false if it failed.
    """
    action_name, section = print_actions(console_printer,
                                         section,
                                         metadata_list)
    if action_name is None:
        return True

    chosen_action = action_dict[action_name]
    try:
        chosen_action.apply_from_section(result,
                                         file_dict,
                                         file_diff_dict,
                                         section)
    except Exception as exception:  # pylint: disable=broad-except
        print_action_failed(log_printer, action_name, exception)
        return False

    return True


def print_segregation(console_printer):
    console_printer.print(format_line(line="",
                                      real_nr="...",
                                      sign="|",
                                      mod_nr="..."),
                          color=FILE_LINES_COLOR)


def show_enumeration(console_printer, title, items, indentation, no_items_text):
    """
    This function takes as input an iterable object (preferably a list or
    a dict). And prints in a stylized format. If the iterable object is
    empty, it prints a specific statement give by the user. An e.g :

    <indentation>Title:
    <indentation> * Item 1
    <indentation> * Item 2

    :param console_printer: Object to print messages on the console.
    :param title:           Title of the text to be printed
    :param items:           The iterable object.
    :param indentation:     Number of spaces to indent every line by.
    :param no_items_text:   Text printed when iterable object is empty.
    """
    if not items:
        console_printer.print(indentation + no_items_text)
    else:
        console_printer.print(indentation + title)
        if isinstance(items, dict):
            for key, value in items.items():
                console_printer.print(indentation + " * " + key + ": " +
                                      value[0])
        else:
            for item in items:
                console_printer.print(indentation + " * " + item)
    console_printer.print()


def show_bear(console_printer, bear, sections, metadata):
    """
    Display all information about a bear.

    :param console_printer: Object to print messages on the console.
    :param bear:            The bear to be displayed.
    :param sections:        The sections to which the bear belongs.
    :param metadata:        Metadata about the bear.
    """
    console_printer.print("{bear}:".format(bear=bear.__name__))
    console_printer.print("  " + metadata.desc + "\n")

    show_enumeration(console_printer,
                     _("Used in:"),
                     sections,
                     "  ",
                     _("No sections."))
    show_enumeration(console_printer,
                     _("Needed Settings:"),
                     metadata.non_optional_params,
                     "  ",
                     _("No needed settings."))
    show_enumeration(console_printer,
                     _("Optional Settings:"),
                     metadata.optional_params,
                     "  ",
                     _("No optional settings."))


def print_bears(console_printer, bears):
    """
    Presents all bears being used in a stylized manner.

    :param console_printer: Object to print messages on the console.
    :param bears:           Its a dictionary with bears as keys and list of
                            sections containing those bears as values.
    """
    if not bears:
        console_printer.print(_("No bears to show."))
    else:
        for bear in sorted(bears.keys(),
                           key=lambda bear: bear.__name__):
            show_bear(console_printer,
                      bear,
                      bears[bear],
                      bear.get_metadata())


def show_bears(local_bears, global_bears, console_printer):
    """
    Extracts all the bears from each enabled section or the sections in the
    targets and passes a dictionary to the show_bears_callback method.

    :param local_bears:         Dictionary of local bears with section names
                                as keys and bear list as values.
    :param global_bears:        Dictionary of global bears with section
                                names as keys and bear list as values.
    :param show_bears_callback: The callback that is used to print these
                                bears. It will get one parameter holding
                                bears as key and the list of section names
                                where it's used as values.
    """
    bears = inverse_dicts(local_bears, global_bears)

    print_bears(console_printer, bears)



from subprocess import Popen, PIPE
import sys

from coalib.bears.LocalBear import LocalBear
from coalib.results.Result import Result, RESULT_SEVERITY
from coalib.settings.Setting import typed_list
from coalib.misc.Shell import escape_path_argument


# We omit this case in our tests for technical reasons
if sys.version_info < (3, 3):  # pragma: no cover
    raise ImportError("PyLint does not support python3 < 3.3")


class PyLintBear(LocalBear):
    def parse_result(self, file, pylint_line):
        parts = pylint_line.split("|", maxsplit=2)

        line_nr = int(parts[0])
        severity = parts[1]
        if severity == "warning":
            severity = RESULT_SEVERITY.NORMAL
        elif severity in ["error", "fatal"]:
            severity = RESULT_SEVERITY.MAJOR
        else:  # convention and refactor
            severity = RESULT_SEVERITY.INFO

        message = parts[2]

        return Result(self.__class__.__name__,
                      message,
                      file,
                      severity,
                      line_nr)

    def run(self,
            filename,
            file,
            pylint_disable: typed_list(str)=("fixme"),
            pylint_enable: typed_list(str)=None,
            pylint_cli_options: str=""):
        '''
        Checks the code with pylint. This will run pylint over each file
        separately.

        :param pylint_disable:     Disable the message, report, category or
                                   checker with the given id(s).
        :param pylint_enable:      Enable the message, report, category or
                                   checker with the given id(s).
        :param pylint_cli_options: Any command line options you wish to be
                                   passed to pylint.
        '''
        command = ('pylint -r n --msg-template="{line}|{category}|'
                   '{msg}. ({msg_id}, {symbol}, {obj})" '
                   + escape_path_argument(filename))
        if pylint_disable:
            command += " --disable=" + ",".join(pylint_disable)
        if pylint_enable:
            command += " --enable=" + ",".join(pylint_enable)
        if pylint_cli_options:
            command += " " + pylint_cli_options

        process = Popen(command,
                        shell=True,
                        stdout=PIPE,
                        stderr=PIPE,
                        universal_newlines=True)
        process.wait()
        current_lines = ""
        for line in process.stdout.readlines():
            if line.startswith("***"):
                continue

            if current_lines != "" and line.split("|", 1)[0].isdigit():
                yield self.parse_result(filename, current_lines)
                current_lines = ""

            current_lines += line

        process.stdout.close()
        process.stderr.close()



import os
import dbus.service

from coalib.output.dbus.DbusApp import DbusApp


class DbusServer(dbus.service.Object):
    interface = "org.coala.v1"

    def __init__(self, bus, path, on_disconnected=None):
        """
        Creates a new DbusServer class which handles the dynamic creation and
        disposal of dbus object-paths for documents and also handles
        information about DbusApplication.

        :param bus:             The dbus bus to which to connect this object
                                path to.
        :param path:            The path in the dbus bus using which apps can
                                communicate.
        :param on_disconnected: This function will be called when the DbusServer
                                has no more applications connected to it.
        """
        dbus.service.Object.__init__(self, bus, path)

        self.apps = {}
        self.__next_app_id = 0
        self.on_disconnected = on_disconnected

        bus.add_signal_receiver(self._on_name_lost,
                                signal_name='NameOwnerChanged',
                                dbus_interface=None,
                                path=None)

    @dbus.service.method(interface,
                         in_signature="s",
                         out_signature="o",
                         sender_keyword="sender")
    def CreateDocument(self, path, sender=None):
        """
        Creates a DbusDocument if it doesn't exist.

        :param path:   The path to the document.
        :param sender: The client who created the dbus request - this is used
                       as the DbusApp's name.
        :return:       a DbusDocument object.
        """
        app = self.get_or_create_app(sender)
        doc = self.get_or_create_document(app, path)
        return doc._object_path

    @dbus.service.method(interface,
                         in_signature="s",
                         out_signature="",
                         sender_keyword="sender")
    def DisposeDocument(self, path, sender=None):
        """
        Disposes a DbusDocument if it exists. Fails silently if it does not
        exist.

        :param path:   The path to the document.
        :param sender: The client who created the dbus request - this is used
                       as the DbusApp's name to search for the document in.
        """
        path = os.path.normpath(path)

        try:
            app = self.apps[sender]
        except KeyError:
            return

        self.dispose_document(app, path)

    def _on_name_lost(self, name, oldowner, newowner):
        if newowner != '':
            return

        self.dispose_app(oldowner)

    def _next_app_id(self):
        self.__next_app_id += 1
        return self.__next_app_id

    def create_app(self, appname):
        """
        Create a new dbus app with the given appname.

        :param appname: The name of the app to be created.
        :return:        a DbusApp object.
        """
        self.apps[appname] = DbusApp(self._next_app_id(), appname)
        return self.apps[appname]

    def get_or_create_app(self, appname):
        """
        Get the dbus app with the given appname. If there does not exist any
        app with the given name, a new app is created and returned.

        :param appname: The name of the app to be created.
        :return:        A DbusApp object.
        """
        try:
            return self.apps[appname]
        except KeyError:
            return self.create_app(appname)

    def dispose_app(self, appname):
        """
        Dispose of the app with the given name. It fails silently if the app
        does not exist. If there are no more apps connected to the server, it
        calls the on_disconnected callback.

        :param appname: The name of the app to dispose of.
        """
        try:
            self.apps.pop(appname)
            if len(self.apps) == 0 and self.on_disconnected:
                self.on_disconnected()
        except KeyError:
            pass

    def create_document(self, app, path):
        """
        Create a new dbus document.

        :param app:  The DbusApp the document is related to.
        :param path: The path to the document to be created.
        :return:     a DbusDocument object.
        """
        doc = app.create_document(path)
        objpath = (self._object_path + "/" + str(app.app_id) +
                   "/documents/" + str(doc.doc_id))
        doc.add_to_connection(self._connection, objpath)

        return doc

    def get_or_create_document(self, app, path):
        """
        Get the dbus document with the given path. If there does not exist any
        document under the DbusApp with the given path, a new document is
        created and returned.

        :param app:  The DbusApp the document is under.
        :param path: The path to the document to be created.
        :return:     A DbusApp object.
        """
        path = os.path.abspath(os.path.expanduser(path))
        try:
            doc = app.docs[path]
        except KeyError:
            doc = self.create_document(app, path)

        return doc

    def dispose_document(self, app, path):
        """
        Dispose of the document with the given path. It fails silently if the
        document does not exist. If there are no more documents in the app,
        the app is disposed.

        :param app:  The DbusApp the document is under.
        :param path: The path to the document.
        """
        doc = app.dispose_document(path)
        if doc != None:
            doc.remove_from_connection()
            if len(app.docs) == 0:
                self.dispose_app(app.name)



from coalib.output.printers.ColorPrinter import ColorPrinter
from coalib.output.printers.LOG_LEVEL import LOG_LEVEL, LOG_LEVEL_COLORS
from coalib.output.printers.LogPrinter import LogPrinter


class ColoredLogPrinter(ColorPrinter, LogPrinter):
    def __init__(self,
                 log_level=LOG_LEVEL.WARNING,
                 timestamp_format="%X",
                 print_colored=True):
        ColorPrinter.__init__(self, print_colored=print_colored)
        LogPrinter.__init__(self, log_level, timestamp_format)

    def _print_log_message(self, prefix, log_message, **kwargs):
        """
        Override this if you want to influence how the log message is printed.
        """
        color = LOG_LEVEL_COLORS[log_message.log_level]
        self.print(prefix, end=" ", color=color, **kwargs)
        self.print(log_message.message, **kwargs)



from functools import total_ordering

from coalib.results.RESULT_SEVERITY import RESULT_SEVERITY
from coalib.results.result_actions.OpenEditorAction import OpenEditorAction


@total_ordering
class Result:
    """
    A result is anything that has an origin and a message.

    Optionally it might affect a file.

    When sorting a list of results with the implemented comparision routines
    you will get an ordering which follows the following conditions,
    while the first condition has the highest priority, which descends to the
    last condition.
    1. Results with no files will be shown first
    2. Results will be sorted by files (ascending alphabetically)
    3. Results will be sorted by lines (ascending)
    4. Results will be sorted by severity (descending, major first, info last)
    5. Results will be sorted by origin (ascending alphabetically)
    6. Results will be sorted by message (ascending alphabetically)
    7. Results will be sorted by debug message (ascending alphabetically)
    """

    def __init__(self,
                 origin,
                 message,
                 file=None,
                 severity=RESULT_SEVERITY.NORMAL,
                 line_nr=None,
                 debug_msg=""):
        """
        :param origin:    Class name or class of the creator of this object
        :param message:   Message to show with this result
        :param file:      The path to the affected file
        :param severity:  Severity of this result
        :param line_nr:   Number of the line which is affected, first line is 1
        :param debug_msg: A message which may help the user find out why
                          this result was yielded.
        """
        if not isinstance(origin, str):
            origin = origin.__class__.__name__

        self.origin = origin
        self.message = message
        self.debug_msg = debug_msg
        self.file = file
        self.line_nr = line_nr
        self.severity = severity

    def __repr__(self):
        return (("<" + Result.__name__ + " object(origin: {origin}, file: "
                 "{file}, line_nr: {linenr}, severity: {severity}, message: "
                 "{msg}) at " + hex(id(self)) + ">")
                    .format(origin=repr(self.origin),
                            file=repr(self.file),
                            linenr=self.line_nr,
                            severity=self.severity,
                            msg=repr(self.message)))

    def __str__(self):
        return ("Result:\n origin: {origin}\n file: {file}\n line nr: "
                "{linenr}\n severity: {severity}\n{msg}"
                .format(origin=repr(self.origin),
                        file=repr(self.file),
                        linenr=self.line_nr,
                        severity=self.severity,
                        msg=repr(self.message)))

    def __eq__(self, other):
        return (isinstance(other, Result) and
                self.origin == other.origin and
                self.message == other.message and
                self.debug_msg == other.debug_msg and
                self.file == other.file and
                self.severity == other.severity and
                self.line_nr == other.line_nr)

    def __lt__(self, other):
        if not isinstance(other, Result):
            raise TypeError("Comparision with non-result classes is not "
                            "supported.")

        # Show elements without files first
        if (self.file is None) != (other.file is None):
            return self.file is None

        # Now either both .file members are None or both are set
        if self.file != other.file:
            return self.file < other.file

        # If we have a line result show results with a lesser line number first
        if self.line_nr is not None and other.line_nr is not None:
            if self.line_nr != other.line_nr:
                return self.line_nr < other.line_nr

        # Both files are equal
        if self.severity != other.severity:
            return self.severity > other.severity

        # Severities are equal, files are equal
        if self.origin != other.origin:
            return self.origin < other.origin

        if self.message != other.message:
            return self.message < other.message

        return self.debug_msg < other.debug_msg

    def get_actions(self):
        """
        :return: All ResultAction classes applicable to this result.
        """
        actions = []
        if self.file is not None:
            actions.append(OpenEditorAction())

        return actions



"""
A Diff result represents a difference for one file.
"""
import difflib

from coalib.results.LineDiff import LineDiff


class ConflictError(Exception):
    pass


class Diff:
    def __init__(self):
        self._changes = {}

    @classmethod
    def from_string_arrays(cls, file_array_1, file_array_2):
        """
        Creates a Diff object from two arrays containing strings.

        If this Diff is applied to the original array, the second array will be
        created.

        :param file_array_1: Original array
        :param file_array_2: Array to compare
        """
        result = cls()

        matcher = difflib.SequenceMatcher(None, file_array_1, file_array_2)
        # We use this because its faster (generator) and doesnt yield as much
        # useless information as get_opcodes.
        for change_group in matcher.get_grouped_opcodes(1):
            for (tag,
                 a_index_1,
                 a_index_2,
                 b_index_1,
                 b_index_2) in change_group:
                if tag == "delete":
                    for index in range(a_index_1+1, a_index_2+1):
                        result.delete_line(index)
                elif tag == "insert":
                    # We add after line, they add before, so dont add 1 here
                    result.add_lines(a_index_1,
                                     file_array_2[b_index_1:b_index_2])
                elif tag == "replace":
                    result.change_line(a_index_1+1,
                                       b_index_1+1,
                                       file_array_2[b_index_1])
                    result.add_lines(a_index_1+1,
                                     file_array_2[b_index_1+1:b_index_2])
                    for index in range(a_index_1+2, a_index_2+1):
                        result.delete_line(index)

        return result

    def _get_change(self, line_nr, min_line=1):
        if not isinstance(line_nr, int):
            raise TypeError("line_nr needs to be an integer.")
        if line_nr < min_line:
            raise ValueError("The given line number is not allowed.")

        return self._changes.get(line_nr, LineDiff())

    def __len__(self):
        return len(self._changes)

    def apply(self, file):
        """
        Applies this diff to the given file.

        :param file: A list of all lines in the file. (readlines) Will not be
                     modified.
        :return:     The modified file.
        """
        result = []
        current_line = 0

        # Note that line_nr counts from _1_ although 0 is possible when
        # inserting lines before everything
        for line_nr in sorted(self._changes):
            result.extend(file[current_line:max(line_nr-1, 0)])
            linediff = self._changes[line_nr]
            if not linediff.delete and not linediff.change and line_nr > 0:
                result.append(file[line_nr-1])
            elif linediff.change:
                result.append(linediff.change[1])

            if linediff.add_after:
                result.extend(linediff.add_after)

            current_line = line_nr

        result.extend(file[current_line:])

        return result

    def __add__(self, other):
        """
        Adds another diff to this one. Will throw an exception if this is not
        possible.
        """
        if not isinstance(other, Diff):
            raise TypeError("Only diffs can be added to a diff.")

        for line_nr in other._changes:
            change = other._changes[line_nr]
            if change.delete is True:
                self.delete_line(line_nr)
            if change.add_after is not False:
                self.add_lines(line_nr, change.add_after)
            if change.change is not False:
                self.change_line(line_nr, change.change[0], change.change[1])

        return self

    def delete_line(self, line_nr):
        """
        Mark the given line nr as deleted. The first line is line number 1.
        """
        linediff = self._get_change(line_nr)
        linediff.delete = True
        self._changes[line_nr] = linediff

    def add_lines(self, line_nr_before, lines):
        """
        Adds lines after the given line number.

        :param line_nr_before: Line number of the line before the additions.
                               Use 0 for insert lines before everything.
        :param lines:          A list of lines to add.
        """
        if lines == []:
            return  # No action

        linediff = self._get_change(line_nr_before, min_line=0)
        if linediff.add_after is not False:
            raise ConflictError("Cannot add lines after the given line since "
                                "there are already lines.")

        linediff.add_after = lines
        self._changes[line_nr_before] = linediff

    def change_line(self, line_nr, original_line, replacement):
        """
        Changes the given line with the given line number. The replacement will
        be there instead.
        """
        linediff = self._get_change(line_nr)
        if linediff.change is not False:
            raise ConflictError("An already changed line cannot be changed.")

        linediff.change = (original_line, replacement)
        self._changes[line_nr] = linediff



import functools
from itertools import combinations

from coalib.misc.StringConverter import StringConverter
from coalib.results.HiddenResult import HiddenResult
from coalib.settings.Setting import typed_ordered_dict
from coalib.bears.GlobalBear import GlobalBear
from bears.codeclone_detection.ClangCountVectorCreator import (
    ClangCountVectorCreator)
from bears.codeclone_detection.ClangCountingConditions import condition_dict
from bears.codeclone_detection.CloneDetectionRoutines import (
    compare_functions,
    get_count_matrices)


# counting_condition_dict is a function object generated by typed_dict. This
# function takes a setting and creates a dictionary out of it while it
# converts all keys to counting condition function objects (via the
# condition_dict) and all values to floats while unset values default to 1.
counting_condition_dict = typed_ordered_dict(
    lambda setting: condition_dict[str(setting).lower()],
    float,
    1)

default_cc_dict = counting_condition_dict(StringConverter(
    """
used: 0,
returned: 2,
is_condition: 0,
in_condition: 1.4,
in_second_level_condition,
in_third_level_condition,
is_assignee: 1.4,
is_assigner: 0.8,
loop_content: 0,
second_level_loop_content,
third_level_loop_content,
is_param: 2,
is_called: 0.2,
is_call_param: 0.8,
in_sum: 1.4,
in_product: 0,
in_binary_operation,
member_accessed: 2"""))


def get_difference(function_pair,
                   count_matrices,
                   average_calculation,
                   reduce_big_diffs):
    """
    Retrieves the difference between two functions using the munkres algorithm.

    :param args: A tuple holding the first function id, the second and the
                 count matrices dictionary holding the count matrices for
                 each function with the function id as key.
    :return:     A tuple containing both function ids and their difference.
    """
    function_1, function_2 = function_pair
    return (function_1,
            function_2,
            compare_functions(count_matrices[function_1],
                              count_matrices[function_2],
                              average_calculation,
                              reduce_big_diffs))


class ClangFunctionDifferenceBear(GlobalBear):
    def run(self,
            counting_conditions: counting_condition_dict=default_cc_dict,
            average_calculation: bool=False,
            reduce_big_diffs: bool=True):
        '''
        Retrieves similarities for code clone detection. Those can be reused in
        another bear to produce results.

        :param counting_conditions:  A comma seperated list of counting
                                     conditions. Possible values are: used,
                                     returned, is_condition, in_condition,
                                     in_second_level_condition,
                                     in_third_level_condition, is_assignee,
                                     is_assigner, loop_content,
                                     second_level_loop_content,
                                     third_level_loop_content, is_param,
                                     in_sum, in_product, in_binary_operation,
                                     member_accessed.
                                     Weightings can be assigned to each
                                     condition due to providing a dict
                                     value, i.e. having used weighted in
                                     half as much as other conditions would
                                     simply be: "used: 0.5, is_assignee".
                                     Weightings default to 1 if unset.
        :param average_calculation: If set to true the difference calculation
                                    function will take the average of all
                                    variable differences as the difference,
                                    else it will normalize the function as a
                                    whole and thus weighting in variables
                                    dependent on their size.
        :param reduce_big_diffs:    If set to true, the difference value of big
                                    function pairs will be reduced. This may be
                                    useful because small functions are less
                                    likely to be clones at the same difference
                                    value than big functions which provide a
                                    better refactoring opportunity for the
                                    user.
        '''
        if not isinstance(counting_conditions, dict):
            self.err("The counting_conditions setting is invalid. Code clone "
                     "detection cannot run.")
            return

        self.debug("Using the following counting conditions:")
        for key, val in counting_conditions.items():
            self.debug(" *", key.__name__, "(weighting: {})".format(val))

        self.debug("Creating count matrices...")
        count_matrices = get_count_matrices(
            ClangCountVectorCreator(list(counting_conditions.keys()),
                                    list(counting_conditions.values()),
                                    self.section["files"].origin),
            list(self.file_dict.keys()),
            lambda prog: self.debug("{:2.4f}%...".format(prog)))

        self.debug("Calculating differences...")

        differences = []
        function_count = len(count_matrices)
        # Thats n over 2, hardcoded to simplify calculation
        combination_length = function_count * (function_count-1) / 2
        partial_get_difference = functools.partial(
            get_difference,
            count_matrices=count_matrices,
            average_calculation=average_calculation,
            reduce_big_diffs=reduce_big_diffs)

        for i, elem in enumerate(
                map(partial_get_difference,
                    [(f1, f2) for f1, f2 in combinations(count_matrices, 2)])):
            if i % 50 == 0:
                self.debug("{:2.4f}%...".format(100*i/combination_length))
            differences.append(elem)

        yield HiddenResult(self.__class__.__name__, differences)
        yield HiddenResult(self.__class__.__name__, count_matrices)



from coalib.misc.Enum import enum
from coalib.misc.i18n import _, N_


RESULT_SEVERITY = enum(N_("INFO"), N_("NORMAL"), N_("MAJOR"))
RESULT_SEVERITY.__str__ = lambda x: _(RESULT_SEVERITY.reverse.get(x, "NORMAL"))
RESULT_SEVERITY_COLORS = {RESULT_SEVERITY.INFO: "green",
                          RESULT_SEVERITY.NORMAL: "yellow",
                          RESULT_SEVERITY.MAJOR: "red"}



from coalib.results.Result import Result
from coalib.bears.LocalBear import LocalBear
from coalib.misc.i18n import _


class KeywordBear(LocalBear):
    def run(self,
            filename,
            file,
            cs_keywords: list,
            ci_keywords: list):
        '''
        Checks the code files for given keywords.

        :param cs_keywords: A list of keywords to search for (case sensitive).
                            Usual examples are TODO and FIXME.
        :param ci_keywords: A list of keywords to search for (case
                            insensitive).
        '''
        for i in range(len(ci_keywords)):
            ci_keywords[i] = ci_keywords[i].lower()

        for line_number, line in enumerate(file):
            found_kws = []
            for keyword in cs_keywords:
                if keyword in line:
                    found_kws.append(keyword)

            for keyword in ci_keywords:
                if keyword in line.lower():
                    found_kws.append(keyword)

            if found_kws != []:
                yield Result(
                    origin=self,
                    message=_("Line contains the following keywords:") +
                            "\n" + ", ".join(found_kws),
                    file=filename,
                    line_nr=line_number + 1)



def close_objects(*objs):
    """
    Determines for all given objects if an object is closable and closes
    it if possible.

    :param objs: The objects to close.
    """
    for obj in objs:
        if isinstance(obj, ClosableObject):
            obj.close()


class ClosableObject:
    """
    Any class deriving from ClosableObject needs to be closed.
    """

    def __init__(self):
        self._closed = False

    def _close(self):
        """
        Override this method to implement own closing behaviour.
        """
        pass

    def close(self):
        """
        Closes any IO Objects this object may hold.
        """
        if not self._closed:
            self._close()
            self._closed = True

    def __del__(self):  # pragma: no cover
        """
        May warn if not closed properly - programmer error!

        Obviously cannot deterministically be tested.
        """
        if not self._closed:
            print("{} needs to be closed "
                  "manually.".format(self.__class__.__name__))
            self.close()



from coalib.misc.Enum import enum
from coalib.misc.i18n import N_

LOG_LEVEL = enum(N_("DEBUG"), N_("WARNING"), N_("ERROR"))
LOG_LEVEL_COLORS = {LOG_LEVEL.ERROR: "red",
                    LOG_LEVEL.WARNING: "yellow",
                    LOG_LEVEL.DEBUG: "green"}



from inspect import ismethod, getfullargspec
from collections import OrderedDict

from coalib.settings.DocumentationComment import DocumentationComment
from coalib.output.printers.ConsolePrinter import ConsolePrinter
from coalib.misc.i18n import _


class FunctionMetadata:
    str_nodesc = _("No description given.")
    str_optional = _("Optional, defaults to '{}'.")

    def __init__(self,
                 name,
                 desc="",
                 retval_desc="",
                 non_optional_params=None,
                 optional_params=None):
        """
        Creates the FunctionMetadata object.

        :param name:                The name of the function.
        :param desc:                The description of the function.
        :param retval_desc:         The retval description of the function.
        :param non_optional_params: A dict containing the name of non optional
                                    parameters as the key and a tuple of a
                                    description and the python annotation. To
                                    preserve the order, use OrderedDict.
        :param optional_params:     A dict containing the name of optional
                                    parameters as the key and a tuple
                                    of a description, the python annotation and
                                    the default value. To preserve the order,
                                    use OrderedDict.
        """
        if non_optional_params is None:
            non_optional_params = OrderedDict()
        if optional_params is None:
            optional_params = OrderedDict()

        self.name = name
        self.desc = desc
        self.retval_desc = retval_desc
        self.non_optional_params = non_optional_params
        self.optional_params = optional_params

    def create_params_from_section(self,
                                   section,
                                   log_printer=ConsolePrinter()):
        """
        Create a params dictionary for this function that holds all values the
        function needs plus optional ones that are available.

        :param section: The section to retrieve the values from.
        :return:        A dictionary. Unfold it with ** to pass it to the
                        function.
        """
        # Import Section only as needed to avoid circular dependency
        from coalib.settings.Section import Section

        if not isinstance(section, Section):
            raise TypeError("The 'section' parameter should be a "
                            "coalib.settings.Section instance.")

        params = {}

        for param in self.non_optional_params:
            dummy, annotation = self.non_optional_params[param]
            params[param] = self._get_param(param,
                                            section,
                                            annotation,
                                            log_printer)

        for param in self.optional_params:
            if param in section:
                dummy, annotation, dummy = self.optional_params[param]
                params[param] = self._get_param(param,
                                                section,
                                                annotation,
                                                log_printer)

        return params

    @staticmethod
    def _get_param(param, section, annotation, log_printer):
        if annotation is None:
            annotation = lambda x: x
        try:
            return annotation(section[param])
        except:
            log_printer.warn(_("Unable to convert {param} to the desired "
                               "data type.").format(param=param))
            return section[param]

    @classmethod
    def from_function(cls, func, omit=()):
        """
        Creates a FunctionMetadata object from a function. Please note that any
        variable argument lists are not supported. If you do not want the
        first (usual named 'self') argument to appear please pass the method of
        an actual INSTANCE of a class; passing the method of the class isn't
        enough. Alternatively you can add "self" to the omit list.

        :param func: The function.
        :param omit: A list of parameter names that are to be ignored.
        :return:     The FunctionMetadata object corresponding to the given
                     function.
        """
        doc = func.__doc__
        if doc is None:
            doc = ""
        doc_comment = DocumentationComment.from_docstring(doc)

        non_optional_params = OrderedDict()
        optional_params = OrderedDict()

        argspec = getfullargspec(func)
        args = argspec.args if argspec.args is not None else ()
        defaults = argspec.defaults if argspec.defaults is not None else ()
        num_non_defaults = len(args) - len(defaults)
        for i, arg in enumerate(args):
            # Implicit self argument or omitted explicitly
            if (i < 1 and ismethod(func)) or arg in omit:
                continue

            if i < num_non_defaults:
                non_optional_params[arg] = (
                    doc_comment.param_dict.get(arg, cls.str_nodesc),
                    argspec.annotations.get(arg, None))
            else:
                optional_params[arg] = (
                    doc_comment.param_dict.get(arg, cls.str_nodesc) + " (" +
                    cls.str_optional.format(str(defaults[i-num_non_defaults]))
                    + ")",
                    argspec.annotations.get(arg, None),
                    defaults[i-num_non_defaults])

        return cls(name=func.__name__,
                   desc=doc_comment.desc,
                   retval_desc=doc_comment.retval_desc,
                   non_optional_params=non_optional_params,
                   optional_params=optional_params)



import os
import tempfile
import subprocess

from coalib.results.Diff import Diff
from coalib.results.result_actions.ApplyPatchAction import ApplyPatchAction

EDITOR_ARGS = {
    "subl": "--wait",
    "gedit": "-s"
}


GUI_EDITORS = ["kate", "gedit", "subl"]


class OpenEditorAction(ApplyPatchAction):
    def apply(self, result, original_file_dict, file_diff_dict, editor: str):
        """
        Open a temporary clone of the file in an editor.

        :param editor: The editor to open the file with.
        """
        filename = result.file
        original_file = original_file_dict[filename]
        diff = file_diff_dict.get(filename, Diff())
        current_file = diff.apply(original_file)

        # Prefix is nice for the user so he has an indication that its the
        # right file he's editing
        temphandle, tempname = tempfile.mkstemp(
            "_" + os.path.basename(filename))
        os.close(temphandle)
        with open(tempname, "w") as temphandle:
            temphandle.writelines(current_file)

        editor_args = [editor, tempname]
        arg = EDITOR_ARGS.get(editor.strip(), None)
        if arg:
            editor_args.append(arg)

        # Dear user, you wanted an editor, so you get it. But do you really
        # think you can do better than we?
        if editor in GUI_EDITORS:
            subprocess.call(editor_args, stdout=subprocess.PIPE)
        else:
            subprocess.call(editor_args)

        with open(tempname) as temphandle:
            new_file = temphandle.readlines()

        os.remove(tempname)

        intermediate_diff = Diff.from_string_arrays(current_file, new_file)
        file_diff_dict[filename] = diff + intermediate_diff

        return file_diff_dict



#!/usr/bin/env python3

from setuptools import setup

from coalib import assert_supported_version
from coalib.misc.i18n import compile_translations
from coalib.misc.StringConstants import StringConstants
from coalib.misc.BuildManPage import BuildManPage


assert_supported_version()


if __name__ == "__main__":
    authors = "Lasse Schuirmann, Fabian Neuschmidt, Mischa Kr\xfcger"
    author_mails = ('lasse.schuirmann@gmail.com, '
                    'fabian@neuschmidt.de, '
                    'makman@alice.de')
    data_files = compile_translations()

    setup(name='coala',
          version=StringConstants.VERSION,
          description='Code Analysis Application (coala)',
          author=authors,
          author_email=author_mails,
          maintainer=authors,
          maintainer_email=author_mails,
          url='http://coala.rtfd.org/',
          platforms='any',
          packages=['bears',
                    'bears.codeclone_detection',
                    'bears.linters',
                    'bears.misc',
                    'bears.spacing',
                    'coalib',
                    'coalib.bearlib',
                    'coalib.bearlib.abstractions',
                    'coalib.bearlib.parsing',
                    'coalib.bearlib.parsing.clang',
                    'coalib.bearlib.spacing',
                    'coalib.bears',
                    'coalib.collecting',
                    'coalib.misc',
                    'coalib.output',
                    'coalib.output.printers',
                    'coalib.output.dbus',
                    'coalib.parsing',
                    'coalib.processes',
                    'coalib.processes.communication',
                    'coalib.results',
                    'coalib.results.result_actions',
                    'coalib.settings'],
          install_requires=["setuptools",
                            "munkres3",
                            "coverage",
                            "pylint",
                            "language-check"],
          package_data={'coalib': ['default_coafile', "VERSION"]},
          license="AGPL v3",
          data_files=data_files,
          long_description="coala is a simple COde AnaLysis Application. Its "
                           "goal is to make static code analysis easy while "
                           "remaining completely modular and therefore "
                           "extendable and language independent. Code analysis"
                           " happens in python scripts while coala manages "
                           "these, tries to provide helpful libraries and "
                           "provides a user interface. Please visit "
                           "http://coala.rtfd.org/ for more information or our"
                           "development repository on "
                           "https://github.com/coala-analyzer/coala/.",
          entry_points={
              "console_scripts": [
                  "coala = coalib.coala:main",
                  "coala-ci = coalib.coala_ci:main",
                  "coala-dbus = coalib.coala_dbus:main"]},
          # from http://pypi.python.org/pypi?%3Aaction=list_classifiers
          classifiers=[
              'Development Status :: 3 - Alpha',

              'Environment :: Console',
              'Environment :: MacOS X',
              'Environment :: Win32 (MS Windows)',
              'Environment :: X11 Applications :: Gnome',

              'Intended Audience :: Science/Research',
              'Intended Audience :: Developers',

              'License :: OSI Approved :: GNU Affero General Public License '
              'v3 or later (AGPLv3+)',

              'Operating System :: OS Independent',

              'Programming Language :: Python :: 3.2',
              'Programming Language :: Python :: 3.3',
              'Programming Language :: Python :: 3.4',

              'Topic :: Scientific/Engineering :: Information Analysis',
              'Topic :: Software Development :: Quality Assurance',
              'Topic :: Text Processing :: Linguistic'],
          cmdclass={'build_manpage': BuildManPage})



from collections import OrderedDict
import sys
import os

from coalib.parsing.LineParser import LineParser
from coalib.settings.Setting import Setting
from coalib.settings.Section import Section


class ConfParser:
    if sys.version_info < (3, 3):  # pragma: no cover
        FileNotFoundError = IOError
    else:
        FileNotFoundError = FileNotFoundError

    def __init__(self,
                 key_value_delimiters=('=',),
                 comment_seperators=('#', ';', '//'),
                 key_delimiters=(',', ' '),
                 section_name_surroundings=None):
        section_name_surroundings = section_name_surroundings or {"[": "]"}

        self.line_parser = LineParser(key_value_delimiters,
                                      comment_seperators,
                                      key_delimiters,
                                      section_name_surroundings)

        # Declare it
        self.sections = None
        self.__rand_helper = None
        self.__init_sections()

    def parse(self, input_data, overwrite=False):
        """
        Parses the input and adds the new data to the existing. If you want to
        catch the FileNotFoundError please take the FileNotFoundError member of
        this object for catching for backwards compatability to python 3.2.

        :param input_data: filename
        :param overwrite:  behaves like reparse if this is True
        :return:           the settings dictionary
        """
        if os.path.isdir(input_data):
            input_data = os.path.join(input_data, ".coafile")

        with open(input_data, "r", encoding='utf-8') as _file:
            lines = _file.readlines()

        if overwrite:
            self.__init_sections()

        self.__parse_lines(lines, input_data)

        return self.sections

    def get_section(self, name, create_if_not_exists=False):
        key = self.__refine_key(name)
        sec = self.sections.get(key, None)
        if sec is not None:
            return sec

        if not create_if_not_exists:
            raise IndexError

        retval = self.sections[key] = Section(str(name),
                                              self.sections["default"])
        return retval

    @staticmethod
    def __refine_key(key):
        return str(key).lower().strip()

    def __add_comment(self, section, comment, origin):
        key = "comment" + str(self.__rand_helper)
        self.__rand_helper += 1
        section.append(Setting(key, comment, origin))

    def __parse_lines(self, lines, origin):
        current_section_name = "default"
        current_section = self.get_section(current_section_name)
        current_keys = []

        for line in lines:
            section_name, keys, value, comment = self.line_parser.parse(line)

            if comment != "":
                self.__add_comment(current_section, comment, origin)

            if section_name != "":
                current_section_name = section_name
                current_section = self.get_section(current_section_name, True)
                current_keys = []
                continue

            if comment == "" and keys == [] and value == "":
                self.__add_comment(current_section, "", origin)
                continue

            if keys != []:
                current_keys = keys

            for section_override, key in current_keys:
                if key == "":
                    continue

                if section_override == "":
                    current_section.add_or_create_setting(
                        Setting(key, value, origin),
                        allow_appending=(keys == []))
                else:
                    self.get_section(
                        section_override,
                        True).add_or_create_setting(
                            Setting(key, value, origin),
                            allow_appending=(keys == []))

    def __init_sections(self):
        self.sections = OrderedDict()
        self.sections["default"] = Section("Default")
        self.__rand_helper = 0



import queue
import threading


class LogPrinterThread(threading.Thread):
    """
    This is the Thread object that outputs all log messages it gets from
    its message_queue. Setting obj.running = False will stop within the next
    0.1 seconds.
    """
    def __init__(self, message_queue, log_printer):
        threading.Thread.__init__(self)
        self.running = True
        self.message_queue = message_queue
        self.log_printer = log_printer

    def run(self):
        while self.running:
            try:
                elem = self.message_queue.get(timeout=0.1)
                self.log_printer.log_message(elem)
            except queue.Empty:
                pass



from coalib.results.result_actions.ResultAction import ResultAction


class ApplyPatchAction(ResultAction):
    def apply(self, result, original_file_dict, file_diff_dict):
        """
        Apply the patch automatically.
        """
        for filename in result.diffs:
            if filename in file_diff_dict:
                file_diff_dict[filename] += result.diffs[filename]
            else:
                file_diff_dict[filename] = result.diffs[filename]

        return file_diff_dict



from coalib.bears.LocalBear import LocalBear
from coalib.results.RESULT_SEVERITY import RESULT_SEVERITY
from coalib.results.Result import Result
from coalib.misc.i18n import _


class LineCountBear(LocalBear):
    def run(self, filename, file, *args):
        '''
        Counts the lines of each file.
        '''
        yield Result(
            origin=self,
            message=_("This file has {count} lines.").format(count=len(file)),
            severity=RESULT_SEVERITY.INFO,
            file=filename)
